/**
 * File for image handling logic.
 */


// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "image.h"
    #include "utility.h"
#endif

#ifndef user_headers_logging
    #define user_headers_logging
    #include "logging.h"
#endif


// Global Variables.
#ifndef LOGGING_CLASS_INIT
    #define LOGGING_CLASS_INIT
    extern Logging logger;
#endif
const Image::PixelStruct Image::pixelBlack = Image::PixelStruct(0);
const Image::PixelStruct Image::pixelWhite = Image::PixelStruct(1);
const Image::PixelStruct Image::pixelRed = Image::PixelStruct(1,0,0);
const Image::PixelStruct Image::pixelGreen = Image::PixelStruct(0,1,0);
const Image::PixelStruct Image::pixelBlue = Image::PixelStruct(0,0,1);
const Image::PixelStruct Image::pixelYellow = Image::PixelStruct(1,1,0);
const Image::PixelStruct Image::pixelPink = Image::PixelStruct(1,0,1);
const Image::PixelStruct Image::pixelLiBlue = Image::PixelStruct(0,1,1);


/**
 * Calculates class input_path and output_path values.
 *
 * :param image_path: Path to an image on the system.
 */
void Image::calculate_paths(std::string image_path) {
    logger.debug("Image::calculate_paths():");
    logger.debug("    image_path: " + image_path);
    logger.debug("");

    // // Calculate output path location.
    output_path = "../output/";
    std::string base_filename = image_path.substr(image_path.find_last_of("/\\") + 1);

    // Grab directory substrings and generate full output path.
    size_t pos = 0;
    std::string token;
    while ((pos = image_path.find("/")) != std::string::npos) {

        // Pull current substring value from input path.
        token = image_path.substr(0, image_path.find("/"));
        image_path.erase(0, pos + 1);

        // Check if directory value is value we care about.
        if ( (token != ".") && (token != "..") && (token != "input") ) {
            output_path += token + "/";
        }

        // Check if directory exists.
        struct stat info;
        if ( (stat(output_path.c_str(), &info) != 0) || (! info.st_mode & S_IFDIR)) {
            // Directory does not exist.
            mkdir(output_path.c_str(), 0777);
        }

    }

    output_path += base_filename;
}


/**
 * Imports image file data, using class variable data.
 */
void Image::import_image_file() {
    logger.debug("Image::import_image_file():");
    logger.debug("");

    logger.info("Attempting to import image data...");
    logger.info("Data from: \"" + input_path + "\"");

    // Open input file.
    std::ifstream input_file;
    input_file.open(input_path, std::ios::binary | std::ios::in);

    if (! input_file.good()) {
        // Failed to open file.
        logger.error("Failed to open file \"" + input_path + "\"");
    } else {
        // File opened. Grab data.
        std::string header;
        input_file >> header;
        // if (strcmp(header.c_str(), "P6") != 0) {
        //     // throw("Can't read input file");
        //     logger.error("Wrong file type, oops: ");
        //     logger.error(header);
        //     logger.error(input_path);
        //     logger.error("");
        // }

        // Save image values to class.
        input_file >> width >> height >> bit_depth;

        // Create and populate image.
        pixel_arr = new PixelStruct[width * height];

        // Skip empty lines in necessary until we get to the binary data.
        input_file.ignore(256, '\n');

        // Read each pixel one by one and convert bytes to floats.
        unsigned char pix[3];
        for (int index = 0; index < width * height; ++index) {
            input_file.read(reinterpret_cast<char *>(pix), 3);
            pixel_arr[index].r = pix[0] / 255.f;
            pixel_arr[index].g = pix[1] / 255.f;
            pixel_arr[index].b = pix[2] / 255.f;
        }
    }

    input_file.close();
}



//region Constructors.

/**
 * Base constructor.
 */
Image::Image() {
    logger.debug("Image::Image Constructor():");
    logger.debug("");

    // Currently test logic.

    // Essentially creates a 10x10 pixel test image.
    input_path = "../input/pbm/400x300/hallway_00.pbm";
    output_path = "../output/test_00.pbm";
    width = 10;
    height = 10;
    bit_depth = 255;
    PixelStruct c = pixelBlack; // Default color. Can select based on below preset colors.

    // Create and populate image.
    pixel_arr = new PixelStruct[width * height];
    for (int index = 0; index < (width * height); index++) {
        pixel_arr[index] = c;
    }
}


/**
 * Constructor that takes input path to image file.
 * Output path is determined based on parsing input path.
 */
Image::Image(std::string user_path) {
    logger.debug("Image::Image Constructor():");
    logger.debug("    user_path: " + user_path);
    logger.debug("");

    logger.info("Constructing image class.");

    // Re-validate path for saving to class.
    path_validator_struct* return_struct = validate_path(user_path);
    int path_type = return_struct->file_type;
    if (return_struct->err_code) {
        logger.error("Error verifying file type: " + return_struct->err_code.message());
        return;
    }
    free_path_validator_struct(return_struct);
    if (path_type == TYPE_FILE) {
        input_path = user_path;
        calculate_paths(user_path);
        import_image_file();
    } else {
        logger.error("Provided path is not an image file. Class initialization failed.");
        return;
    }

    logger.info("Image class created.");
}


/**
 * Deconstructor.
 */
Image::~Image() {
    logger.debug("Deconstructing image class");

    // Deallocate memory objects.
    delete[] pixel_arr;
}

//endregion Constructors.


//region Public Methods.

/**
 * Displays general properties of image held in class.
 */
void Image::display_properties() {
    logger.debug("Image::display_properties():");
    logger.debug("");

    logger.info("");
    logger.info("input file path: " + input_path);
    logger.info("output file path: " + output_path);
    logger.info("width: " + std::to_string(width));
    logger.info("height: " + std::to_string(height));
    logger.info("bit depth: " + std::to_string(bit_depth));
    logger.info("");
}


/**
 * Saves class image data to output file.
 */
void Image::save() {
    logger.debug("Image::save():");
    logger.debug("");

    logger.debug("Attempting to save image to \"" + output_path + "\".");

    // Open output file stream.
    std::ofstream output_file;

    // Attempt to save file data.
    try {
        // Attempt to open file.
        output_file.open(output_path, std::ios::binary | std::ios::out);

        // Check if opened successfully.
        if (output_file.fail()) {
            // Failed to open.
            logger.error("Failed to open file \"" + output_path + "\"");
        } else {
            // Opened successfully.

            // Write image "header" data.
            output_file << "P6\n" << width << " " << height << "\n255\n";

            // Write image pixel data.
            unsigned char r, g, b;
            for (int index = 0; index < (width * height); index++) {
                r = static_cast<float>(std::min(1.f, pixel_arr[index].r) * 255);
                g = static_cast<float>(std::min(1.f, pixel_arr[index].g) * 255);
                b = static_cast<float>(std::min(1.f, pixel_arr[index].b) * 255);
                output_file << r << g << b;
            }

        }

        // Close file stream.
        output_file.close();
        logger.info("Image saved.");

    } catch (const char* err) {
        logger.error("Failed to save image: " + std::string(err));
        output_file.close();
    }
}

//endregion Public Methods.
